"""
Author      :   Pablo Luesia Lahoz
E-Mail      :   pluesia at unizar.es
File        :   pf_solver.py
Description :   Solves the NLOS problem using the phasor fields approximation,
                applying Rayleigh-Sommerfeld propagation to fill the given 
                voxelization
"""
import numpy as np
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from tqdm import tqdm
from typing import Union, Tuple

from tal.reconstruct.pf.propagator import Propagator, RSD_propagator
from tal.reconstruct.pf.propagator import RSD_parallel_propagator

__epsilon = 1e-5

def pulse(delta_t: float, n_w: int, lambda_c: float, cycles: float
          ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Generate a gaussian pulse in frequency space. It assume the data 
    bins are equally space in time. Returns the pulse in frequency,
    the wavelengths of the frequencies extracted from the data, and
    the indices to contain the 99.74% of the frequencies given the 
    centered frequency with lambda_c wavelength, repeated cycles times
    @param  delta_t     : Time spacing between bins
    @param  n_w         : Number of frequencies extracted with numpy fft
    @param  lambda_c    : Central wavelength of the pulse
    @param  cycles      : Number of pulses to consider of the data
    @return             : The values of the frequency pulse, the wave-
                            lengths of the frequencies to return and the 
                            indices to contain the 99.74% distribution of the
                            frequencies

    """
    sigma = cycles*lambda_c / 6             # Gaussian sigma
    w_c = 1 / lambda_c                      # Central frequency
    w = np.fft.fftfreq(n_w, delta_t)        # All frequencies
    # To control division by 0
    w[0] = 1e-30
    wavelengths = 1 / w                     # All wavelengths
    # Correct 0 and inf values
    wavelengths[0] = np.inf
    w[0] = 0.0
    # Frequency distances to central frequency in rad/s
    delta_w = 2*np.pi*(w - w_c)
    # Gaussian pulse for reconstruction
    freq_pulse = np.exp(-sigma**2*delta_w**2/2)*np.sqrt(2*np.pi)*sigma

    # Central freq and sigma to k up to numpy fft
    central_k = delta_t * n_w * w_c
    sigma_k = delta_t * n_w / (2*np.pi*sigma)

    # Interest indicies in range [mu-3sigma, mu+3sigma]
    min_k = np.round(central_k - 3*sigma_k)
    max_k = np.round(central_k + 3*sigma_k)
    # Indices in the gaussian pulse
    indices = np.arange(min_k, max_k+1, dtype=int)

    return (freq_pulse, wavelengths, indices)


def H_to_fH(H: np.ndarray, t_bins: np.ndarray,  lambda_c: float,
            cycles: float
            ) -> Tuple[np.ndarray, np.ndarray, np.ndarray,
                       Union[np.ndarray, None]]:
    """
    Transforms the data H in time, to the desired Fourier space frequencies
    @param H        : The impulse response of the scene
    @param t_bins   : Time of each temporal bin
    @param lambda_c : The wavelength of the central frequency for the virtual
                      illumination pulse
    @param cycles   : Number of cycles of the virtual illumination pulse
    @return         : Tuple containting first the impulse response in fourier
                      domain for those non zero values, second the wavelengths
                      associated with each fourier component, third the pulse
                      value, and fourth the indices of the pulse according the
                      fft module of numpy
    """
    if cycles == 0:
        # Fourier term function
        fourier_term = np.exp(-2 * np.pi * t_bins / lambda_c * 1j)
        fH = np.array([np.sum(H*fourier_term.reshape((-1,)+(1,)*(H.ndim - 1)),
                              axis=0)])
        wl = np.array([lambda_c])
        f_pulse = np.ones(1)
        significant_idx = None
    else:
        fH_all = np.fft.fft(H, axis=0)
        f_pulse, wv_all, significant_idx = pulse(t_bins[1], fH_all.shape[0],
                                                 lambda_c, cycles)
        wl = wv_all[significant_idx]
        fH = fH_all[significant_idx]

    return (fH, wl, f_pulse, significant_idx)


# Auxiliary function to transform fourier to prime planes,
# given the pulse parameters
def fIz2Iz(fIz: np.ndarray, f_pulse: np.ndarray, significant_idx: np.ndarray,
           idx_t: np.ndarray = np.array([0])) -> np.ndarray:
    """
    Transforms the plane reconstruction I in Fourier domain to time domain
    @param fIz              : A plane reconstruction of the scene by frequency
                              domain
    @param f_pulse          : Virtual ilumination pulse in frequency space
    @param significant_idx  : Indices with enough significance in the virtual
                              illumination pulse to consider, in the numpy fft
                              of the original signal
    @param idx_t            : Evaluation in fft indices of the time
    @return                 : The plane reconstruction of the scene in time 
                              domain, evaluated at the time given by idx_t
    """
    # Shape of the volume
    fIz_shape = tuple(np.array(fIz.shape)[1:])
    # All frequencies to use
    nw_all = f_pulse.shape[0] + significant_idx.shape[0] - 1
    nw_sig_max = np.max(significant_idx)
    # All no indicated frequencies values are 0
    all_freq = np.zeros((nw_sig_max + 1), dtype=np.complex128)
    Iz = np.zeros(idx_t.shape + fIz_shape, dtype=np.complex128)

    # Index along points of fIz
    for it in np.ndindex(fIz_shape):
        exp_it = (np.s_[:],) + it
        fIp = fIz[exp_it]
        # Fill the significant data weighted with the illumination
        # pulse value
        all_freq[significant_idx] = fIp*f_pulse[significant_idx]
        # Inverse FFT to return the data, evaluated at time 0
        Iz[exp_it] = np.fft.ifft(all_freq, n=nw_all)[idx_t]

    return Iz


def fI_to_I(fI: np.ndarray, f_pulse: np.ndarray, sig_idx: np.ndarray,
            idx_t: np.ndarray = np.array([0]), n_threads: int = 1,
            desc: str = "Fourier to time reconstruction"
            ) -> np.ndarray:
    """
    Transforms the volume reconstruction I in Fourier domain to time domain
    executed in parallel
    @param fI           : A volume reconstruction of the scene by frequency 
                          domain (in the first axis)
    @param f_pulse      : Pulse used for the reconstruction in Fourier domain
    @param sig_idx      : Iff not None, indices that correspond in the f_pulse
                         to the most significant values of the pulse
    @param idx_t        : Evaluation in fft indices of the time
    @param n_threads    : Number of threads to use for the function
    @param desc         : Description to show in the progress bar. If None, it
                          shows no progress bar 
    @return             : The volume reconstruction of the scene in time 
                          domain, evaluated in t. By default t=[0]
    """
    assert fI.ndim >= 2 and fI.ndim <= 4, "This data type is not supported"
    if sig_idx is None:    # Single frequency
        # No need to convert the data
        return fI[0]
    else:                   # Multiple frequency
        assert fI.shape[0] == len(sig_idx), "Different number of frequencies"
        with ProcessPoolExecutor(max_workers=n_threads) as ex:
            fIz2Iz_partial = partial(fIz2Iz,
                                     f_pulse=f_pulse,
                                     significant_idx=sig_idx,
                                     idx_t=idx_t)
            units = {1: 'points', 2: 'rows', 3: 'planes'}
            return np.array(
                list(
                    tqdm(
                        ex.map(fIz2Iz_partial, fI.swapaxes(0, 1)),
                        desc=desc,
                        disable=desc is None,
                        unit=units.get(fI.ndim - 1, ' '),
                        total=fI.shape[1]))).swapaxes(0, 1)


def propagator(P: np.ndarray, V: np.ndarray, wl: np.ndarray) -> Propagator:
    """
    Returns the proper RSD propagator given P, V and wl
    @param P    : Origin points to performance the propagator. It can be an 
                  array or a matrix of 3d coordinates
    @param V    : Destination points to propagate. It can be an array, a
                  matrix or a 3d volume of 3d coordinates
    @param wl   : Array of wavelengths to propagate in metters
    @return     : Propagator object
    """
    assert P.ndim >= 2 and P.ndim <= 3 and P.shape[-1] == 3, \
        "Unsupported data format"
    assert V.ndim >= 2 and V.ndim <= 4 and V.shape[-1] == 3, \
        "Unsupported data format"
    planes = True
    if P.ndim == 3:
        if V.ndim == 4:
            target_plane = V[0]
        elif V.ndim == 3:
            target_plane = V
        else:
            planes = False
    else:
        planes = False

    if planes and target_plane.shape == P.shape \
            and target_plane.shape >= (2, 2, 3) and P.shape >= (2, 2, 3) \
            and __parallel(target_plane, P):
        # Check the planes are in front
        v1 = target_plane[0, 0] - P[0, 0]
        v2 = target_plane[-1, 0] - P[-1, 0]
        v3 = target_plane[0, -1] - P[0, -1]
        v4 = target_plane[-1, -1] - P[-1, -1]
        if np.linalg.norm(v1-v2) < __epsilon \
            and np.linalg.norm(v3-v2) < __epsilon \
            and np.linalg.norm(v4-v3) < __epsilon:
            # Parallel propagator
            return RSD_parallel_propagator(P, V, wl)

    return RSD_propagator()


def reconstruct(H:  np.ndarray, t_bins:  np.ndarray, S:  np.ndarray,
                L:  np.ndarray,  V: np.ndarray, lambda_c: float = 6,
                cycles: float = 4, res_in_freq: bool = False,
                n_threads: int = 1, verbose: int = 0) -> np.ndarray:
    """
    Returns a NLOS solver object based on Phasor Fields
    @param H            : Array with the impulse response of the scene by time,
                          camera points and laser points
    @param t_bins       : Temporal stamp for each bin 
    @param S            : Coordinates for each sensor point in z y x in the 
                          relay wall
    @param L            : Coordinates for each light source point in z y x in
                          the relay wall
    @param V            : Voxelization of the reconstructions
    @param lambda_c     : Central wavelength for the virtual illumination 
                          gaussian pulse for the reconstruction
    @param cycles       : Number of cycles of the virtual illumination pulse
    @param res_in_freq  : Iff true, returns the data in the fourier domain 
                          (with the first axis for each frequency). It returns
                          the result in time domain 0 otherwise
    @param n_threads    : Number of threads to use on the reconstruction
    @param verbose      : Set the level of verbose:
                            - 0: prints nothing
                            - 1: informs about the start of the proceedings
                            - 2: 1 and gives information about the pulse, and
                                 shows progress bars for the reconstructions
    @return             : A Phasor Fields approach reconstruction of a NLOS 
                          problem
    """
    assert (V.ndim <= 4 and V.ndim >= 2) and V.shape[-1] == 3, \
        "reconstruct does not support V data"
    assert (S.ndim == 3 or S.ndim == 2) and S.shape[-1] == 3,\
        "reconstruct does not support this S data"
    assert (L.ndim == 3 or L.ndim == 2) and L.shape[-1] == 3,\
        "reconstruct does not support this L data"

    # Reshape all the data to match all the data
    S_r = S
    L_r = L
    if S_r.ndim == 2:
        S_r = S_r.reshape(-1, 1, 3)
    if L_r.ndim == 2:
        L_r = L_r.reshape(-1, 1, 3)
    H_r = __H_format(H, S, L)

    units = {1: 'points', 2: 'rows', 3: 'planes'}
    unit = units.get(V.ndim-1, ' ')

    # Time domain impulse response to Fourier
    __v_print(f"Generating virtual illumination pulse:\n" +
              f"\tCentral wavelength: {lambda_c} m\n" +
              f"\t{cycles} cycles\n...", 1, verbose)

    f_H, wv, f_pulse, sig_idx = H_to_fH(H_r, t_bins, lambda_c, cycles)
    del H_r

    __v_print(f"Done. {len(wv)} frequencies to use", 1, verbose)

    __v_print("Generating propagator from sensors...", 2, verbose)
    propagator_S = propagator(S_r, V, wv)
    __v_print("Done", 2, verbose)
    __v_print("Generating propagator from lights...", 2, verbose)
    propagator_L = propagator(L_r, V, wv)
    __v_print("Done", 2, verbose)
    __v_print(f"Propagating with {n_threads} threads...", 1, verbose)

    # Propagate using multithreading
    propagate_partial = partial(__propagate, propagator_S, propagator_L, f_H,
                                S_r, L_r, wv)
    with ThreadPoolExecutor(max_workers=n_threads) as executor:
        fI = np.array(
            list(tqdm(
                executor.map(propagate_partial, V),
                desc=unit + ' reconstructed',
                disable=verbose < 3,
                unit=unit,
                total=V.shape[0]
            )))
    del S_r, L_r, f_H
    __v_print("Done", 1, verbose)

    if not res_in_freq and sig_idx is not None:     # Result in time domain
        __v_print(f"Transforming from Fourier to time domain with {n_threads}"
                  + " threads...", 1, verbose)
        if verbose >= 3:
            desc = unit + ' to time domain'
        else:
            desc = None
        # Transform the data to Fourier domain
        I = fI_to_I(fI.swapaxes(0, 1), f_pulse, sig_idx, n_threads=n_threads,
                    desc=desc)[0]
        __v_print("Done", 1, verbose)
        return I
    elif not res_in_freq and sig_idx is None:
        return fI[:, 0, ...]
    else:
        return fI


###############################################################################
#                       Auxiliary methods and functions                       #
###############################################################################

# Propagate given the propagators, the impulse response in fourier
# f_H, the sensor points S, the light points L, the target coordinates V_z,
# with frecuency wavelengths wv
def __propagate(propagator_S, propagator_L, f_H, S, L, wv, V):
    # Propagate from sensors
    fI_s = propagator_S.propagate(f_H, S, V, wv, P_axis=(1, 2))
    # Propagate from Lights
    v_axis = tuple(np.arange(-V.ndim+1, 0))
    if len(v_axis) == 0:
        v_axis = (-1,)
    fI = propagator_L.propagate(fI_s, L, V, wv, P_axis=(1, 2), V_axis=v_axis)
    return fI


# Prints only iff threshold =< given_verbose
def __v_print(msg, threshold, given_verbose):
    if threshold <= given_verbose:
        print(msg)


# Update the format of H to the dimensions of S and L
def __H_format(H: np.ndarray, S: np.ndarray, L: np.ndarray):
    n_bins_shape = (H.shape[0], )

    rec_S = S.ndim == 3
    if rec_S:
        S_shape = (S.shape[0], S.shape[1])
    else:
        S_shape = (S.shape[0], 1)

    rec_L = L.ndim == 3
    if rec_L:
        L_shape = (L.shape[0], L.shape[1])
    else:
        L_shape = (L.shape[0], 1)

    return H.reshape(n_bins_shape + S_shape + L_shape)


# Return true if the two planes are parallel
def __parallel(S1, S2):
    # Vectors in plane 1
    p1 = S1[0, 0]
    p2 = S1[0, -1]
    p3 = S1[-1, 0]
    v1 = p3 - p1
    v2 = p2 - p1
    # Normal to the plane
    n1 = np.cross(v1, v2)

    # Vectors in plane 2
    p1 = S2[0, 0]
    p2 = S2[0, -1]
    p3 = S2[-1, 0]
    v1 = p3 - p1
    v2 = p2 - p1
    # Normal to the plane
    n2 = np.cross(v1, v2)

    # Return similar to 0 with error
    return np.linalg.norm(np.cross(n1, n2)) <= __epsilon



from __future__ import annotations

from . import io, plot, reconstruct
from . import enums, callbacks
from tal.config import set_resources, ResourcesConfig as resources
from tal.log import set_log_level, LogLevel

__version__ = '0.16.0'



"""
tal.reconstruct.bp
==================

Reconstruction using the backprojection algorithm.
See "Recovering three-dimensional shape around a corner using ultrafast time-of-flight imaging."

This implementation is particularly well suited when:

- The 3D volume that you want to reconstruct is *not* coplanar to the relay wall
and/or
- You want a time-gated reconstruction, not a time-resolved reconstruction
and/or
- You have a projector camera system and only want to focus light to a single point

If that is not your case, you might want to check the tal.reconstruct.pf_dev module.

Filtered backprojection is available in the fbp submodule, or by pre-filtering (see tal.reconstruct.filter_H).

Does _NOT_ attempt to compensate effects caused by attenuation:
    - cos decay of {sensor|laser}_grid_normals are ignored
    - cos decay of hidden geometry is ignored
    - 1/d^2 decay
"""

from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
import numpy as np


def solve(data: NLOSCaptureData,
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
          camera_system: CameraSystem = CameraSystem.DIRECT_LIGHT,
          projector_focus: NLOSCaptureData.Array3 = None,
          compensate_invsq: bool = False,
          progress: bool = True) -> NLOSCaptureData.SingleReconstructionType:
    """
    See module description of tal.reconstruct.bp

    data
        See tal.io.read_capture

    volume_xyz
        Multi-dimensional array with XYZ coordinates of the volume voxels.
        See tal.enums.VolumeFormat for possible input formats.
        See e.g. tal.reconstruct.get_volume_min_max_resolution for utilities to generate this array.

    volume_format
        See tal.enums.VolumeFormat

    camera_system
        See tal.enums.CameraSystem

    projector_focus
        Setting that changes how the virtual light is focused onto the hidden volume.
        'None' focuses the virtual light at the same points as the virtual camera (confocal camera)
            This is the behaviour of most imaging algorithms,
            especially the tal.enums.CameraSystem.CONFOCAL_TIME_GATED camera.
        When projector_focus = [x, y, z] if you have multiple laser points in your data,
            the illumination will be focused towards the point [x, y, z].
            This behaviour is especially useful for tal.enums.CameraSystem.PROJECTOR_CAMERA.
        In the pf_dev module you can set projector_focus = volume_xyz and that will yield
            a NLOSCaptureData.ExhaustiveReconstructionType with all possible projector_focus points.

    compensate_invsq
        If True, the inverse square falloff of light is compensated for, i.e., objects further away
        from the relay wall will appear brighter in the reconstruction.

    progress
        If True, shows a progress bar with estimated time remaining.
    """
    from tal.reconstruct.util import convert_to_N_3, convert_reconstruction_from_N_3
    H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3, _, __ = \
        convert_to_N_3(data, volume_xyz, volume_format)

    from tal.reconstruct.bp.backprojection import backproject
    reconstructed_volume_n3 = backproject(
        H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3, volume_xyz.shape[:-1],
        camera_system, data.t_accounts_first_and_last_bounces,
        data.t_start, data.delta_t, data.is_confocal(),
        projector_focus,
        data.laser_xyz, data.sensor_xyz,
        compensate_invsq=compensate_invsq,
        progress=progress)

    return convert_reconstruction_from_N_3(data, reconstructed_volume_n3, volume_xyz, volume_format, camera_system)



"""
tal.reconstruct.pf
==================

Reconstruction using the phasor fields framework.
See "Non-Line-of-Sight Imaging using Phasor Field Virtual Wave Optics."

Does _NOT_ attempt to compensate effects caused by attenuation:
    - cos decay i.e. {sensor|laser}_grid_normals are ignored
    - 1/d^2 decay
"""

from tal.io.capture_data import NLOSCaptureData
from tal.reconstruct.pf.propagator import Propagator
from typing import Any, Tuple
import numpy as np


def to_Fourier(data: NLOSCaptureData,
               wavefactor: float,
               pulse_cycles: float) -> Tuple[np.ndarray, np.ndarray, Any]:
    """
    Given the data, and the pulse defined by the wavefactor and cycles, return
    the data in the Fourier domain and the wavelengths associated to each 
    component
    @param data         : NLOS captured data
    @param wavefactor   : Wavefactor of the central frequency of the virtual
                          gaussian illumination pulse. The wavelength value is
                          wavefactor times sensor spacing of the relay wall
    @param pulse_cycles : Cycles of the virtual gaussian illumination pulse.
                          It defines the number of frequencies to use. If it 
                          is set to 0, it will use only a frequency. Otherwise
                          the lower the more frequencies
    @return             : A tuple, with the first element being the different
                          Fourier components of H given the pulse, the second
                          the different wavelengths associated to each 
                          component, and the third auxiliary information to
                          performance the inverse operation
    """
    from tal.reconstruct.pf.pf_solver import H_to_fH
    # Data extraction
    H = data.H
    # Time extraction
    if data.t_start is not None:
        # Correct delayed data
        t_pad = int(np.round(data.t_start / data.delta_t))
        H = np.pad(H, ((t_pad, 0), (0, 0), (0, 0)), 'minimum')
    T = data.delta_t*np.arange(H.shape[0])
    # Calculate the wavelength of the central frequency of the pulse
    # FIXME(pablo): Minimun distance assumes ordered points
    reshaped_S = data.sensor_grid_xyz.reshape(-1, 3)
    lambda_c = wavefactor * np.linalg.norm(reshaped_S[0] - reshaped_S[1])
    # Transform to Fourier domain and extract the wavelengths and auxiliary
    # parameters
    fH, wl, f_pulse, sig_idx = H_to_fH(H, T, lambda_c, pulse_cycles)
    return fH, wl, (f_pulse, sig_idx, data.delta_t)


def to_time(fourier_comp: np.ndarray,
            aux_param: Any,
            n_threads: int = 1,
            t_eval: np.ndarray = None) -> np.ndarray:
    """
    Given the Fourier components it transforms the data to time domain, and
    returns the evaluations at the given time
    @param fourier_comp : Fourier components data to transform to time domain
    @param aux_param    : Auxiliary parameters used to transform it back to
                          the original time domain
    @param n_threads    : Number of threads to use for the transformation
    @param t_eval       : Array of time values to evaluate the transformation
    @return             : Array of results of the evaluations for the given
                          time stamps at the data in time. If there is only a
                          frequency, it return the first element of the 
                          array
    """
    from tal.reconstruct.pf.pf_solver import fI_to_I
    # Extract the auxiliary params
    f_pulse, sig_idx, delta_t = aux_param
    # Approximate the time stamps to the fft index
    if t_eval is None:
        t_eval_it = np.array([0])
    else:
        t_eval_it = np.round(t_eval / delta_t).astype(int)
    # Return the result in time domain
    return fI_to_I(fourier_comp, f_pulse, sig_idx, t_eval_it, n_threads)


def get_propagators(data: NLOSCaptureData,
                    voxels: np.ndarray,
                    wl: np.ndarray) -> Tuple[Propagator, Propagator]:
    """
    Return the propagators from the sensor points, and from the light sources
    of the data to the voxels
    @param data     : NLOS captured data 
    @param voxels   : Target voxels to obtain the propagators
    @param wl       : Array of wavelengths that will propagate
    @return         : Tuple of propagators, being the first one the sensor
                      points to voxels propagator, and the second the light
                      sources to the voxels propagator
    """
    from tal.reconstruct.pf.pf_solver import propagator
    S = data.sensor_grid_xyz
    L = data.laser_grid_xyz
    return (propagator(S, voxels, wl), propagator(L, voxels, wl))


def solve(data: NLOSCaptureData, wavefactor: float, wave_cycles: float,
          volume: NLOSCaptureData.VolumeXYZType = None, res_in_freq=False,
          verbose: int = 1, n_threads: int = 1) -> np.ndarray:
    """
    Reconstruct a NLOS captured data with a gaussian pulse in the volume
    using phasor fields and the RSD propagation.
    @param data         : NLOS capture data at the relay wall
    @param wavefactor   : Central wavefactor of the virtual illumination pulse
                          used for the reconstruction. The wavelength will be 
                          wavefactor*delta_x (distance between relay wall 
                          sensor points)
    @param wave_cycles  : Number of central wavelength cycles of the virtual
                          illumination pulse
    @param volume       : Desired volume in the 3d space to reconstruct from 
                          the relay wall
    @param res_in_freq  : If True, return the result by frequencies, without 
                          convining the used frequencies into the time domain.
                          If False, return the result in time domain, evaluated
                          in t=0
    @param verbose      : It can be used to select from 0 to 3 levels of 
                          information printed by screen. Iff 0 prints nothing,
                          if 3 prints the whole information
    @param n_threads    : Number of threads to use for the reconstruction. By
                          default is set to 1
    @return             : Reconstruction from the relay wall data to the given
                          volume, in the given form.    
    """
    from tal.reconstruct.pf.pf_solver import reconstruct
    # Data extraction
    H = data.H
    # Time extraction
    if data.t_start is not None:
        # Correct delayed data
        t_pad = int(np.round(data.t_start / data.delta_t))
        H = np.pad(H, ((t_pad, 0), (0, 0), (0, 0)), 'minimum')
    T = data.delta_t*np.arange(H.shape[0])

    # TODO(pablo) correct volume
    V = volume
    S = data.sensor_grid_xyz
    L = np.zeros((1, 3), float)

    # FIXME(pablo): Minimun distance assumes ordered points
    reshaped_S = S.reshape(-1, 3)
    wavelength = wavefactor*np.linalg.norm(reshaped_S[0] - reshaped_S[1])

    return reconstruct(H, T, S, L, V, wavelength, wave_cycles, res_in_freq,
                       n_threads, verbose)



from tal.io.capture_data import NLOSCaptureData
from tal.enums import HFormat
from tal.util import SPEED_OF_LIGHT
from tal.log import TQDMLogRedirect
from typing import Union
from nptyping import NDArray, Shape
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.widgets import Slider, Button
import numpy as np
from tqdm import tqdm


def plot_amplitude_phase(image: NLOSCaptureData.SingleReconstructionType, title: str = ''):
    if image.squeeze().ndim != 2:
        raise AssertionError('Image must be 2D')
    image_show = image.squeeze().T
    fig, ax = plt.subplots(1, 2, figsize=(12, 5))
    mappable = ax[0].imshow(np.abs(image_show), cmap='hot')
    fig.colorbar(mappable, ax=ax[0])
    mappable = ax[1].imshow(np.angle(image_show), cmap='seismic',
                            vmin=-np.pi, vmax=np.pi)
    fig.colorbar(mappable, ax=ax[1])
    ax[0].set_title('Amplitude')
    ax[0].set_xlabel('x')
    ax[0].set_ylabel('y')
    ax[1].set_title('Phase')
    ax[1].set_xlabel('x')
    ax[1].set_ylabel('y')
    fig.suptitle(title)
    plt.show()


def plot_xy_grid(data: Union[NLOSCaptureData, NLOSCaptureData.HType],
                 size_x: int, size_y: int,
                 t_start: int, t_end: int, t_step: int):
    def t_to_time(t): return f'Bin #{t_start + t * t_step}'
    if isinstance(data, NLOSCaptureData):
        assert data.H_format == HFormat.T_Sx_Sy, \
            'plot_xy_grid does not support this data format'
        txy = data.H
        if data.t_start is not None and data.delta_t is not None:
            def t_to_time(
                t): return f'Bin #{(t_start or 0) + t * t_step}, {(data.t_start + t * data.delta_t) * 1e12 / SPEED_OF_LIGHT:.0f} ps'
    else:
        assert data.ndim == 3 and data.shape[1] == data.shape[2], \
            'plot_xy_grid does not support this data format'
        txy = data
    txy = txy[t_start:t_end:t_step, ...]
    nt = txy.shape[0]
    step = 1
    plot_size = size_x * size_y
    while nt // step > plot_size:
        step *= 2
    txy_min, txy_max = np.min(txy), np.max(txy)
    fig, axs = plt.subplots(size_y, size_x)

    for i in tqdm(range(plot_size), file=TQDMLogRedirect()):
        t_bin = i * step
        image = txy[t_bin]
        row = i // size_x
        col = i % size_x
        mappable = axs[row, col].imshow(image.astype(
            np.float32), cmap='jet', vmin=txy_min, vmax=txy_max)
        fig.colorbar(mappable, ax=axs[row, col])
        axs[row, col].axis('off')
        axs[row, col].set_title(t_to_time((t_start or 0) + t_bin * t_step))

    plt.tight_layout()
    plt.show()


def plot_3d_interactive_axis(xyz: np.ndarray, focus_slider: NDArray[Shape['T'], NLOSCaptureData.Float],
                             axis: int, title: str, slider_title: str,
                             slider_unit: str, cmap: str = 'hot',
                             xlabel: str = '', ylabel: str = ''):
    assert xyz.ndim == 3, 'Unknown H_Format to plot'
    assert axis < 3, f'Data only has 3 dims (given axis={axis})'
    assert xyz.shape[axis] == len(focus_slider), \
        'The slider and the data have different lengths'
    # Move the axis, so the interactive axis is at 0
    xyz_p = np.moveaxis(xyz, axis, 0)
    max_idx = xyz_p.shape[0] - 1
    v_min = np.min(xyz)
    v_max = np.max(xyz)

    # Plot the first figure
    fig = plt.figure()
    ax = plt.axes([0.2, 0.2, 0.7, 0.7])
    cax = plt.axes([0.85, 0.2, 0.03, 0.7])
    colorbar = None
    scale = 'linear'

    def set_ax_common():
        nonlocal colorbar
        fig.suptitle(title)
        if colorbar is not None:
            cax.cla()
        colorbar = fig.colorbar(img, cax=cax, shrink=0.5)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_xticks([])
        ax.set_yticks([])

    def scale_linear(_):
        nonlocal img, scale
        scale = 'linear'
        ax.cla()
        img = ax.imshow(xyz_p[current_idx], cmap=cmap, vmin=v_min, vmax=v_max)
        set_ax_common()

    def scale_log(_):
        nonlocal img, scale
        scale = 'log'
        ax.cla()
        img = ax.imshow(xyz_p[current_idx] + 1, cmap=cmap,
                        norm=LogNorm(vmin=v_min + 1, vmax=v_max))
        set_ax_common()

    current_idx = 0
    img = None
    scale_linear(None)

    ax_slider = plt.axes([0.25, 0.1, 0.5, 0.03])
    slider = Slider(
        ax=ax_slider,
        label=slider_title,
        valmin=0,
        valmax=len(focus_slider) - 1,
        valinit=0,
        valstep=1.0,
        orientation='horizontal')
    ax_text = plt.axes([0.25, 0.05, 0.5, 0.03])
    ax_text.axis('off')
    text = ax_text.text(0, 0, f'{focus_slider[0]:.2f} {slider_unit}')

    ax_button_linear = plt.axes([0.05, 0.4, 0.1, 0.05])
    button_linear = Button(ax_button_linear, 'Linear')
    button_linear.on_clicked(scale_linear)
    ax_button_log = plt.axes([0.05, 0.45, 0.1, 0.05])
    button_log = Button(ax_button_log, 'Log')
    button_log.on_clicked(scale_log)

    def cmap_hot(_):
        nonlocal cmap
        cmap = 'hot'
        if scale == 'linear':
            scale_linear(None)
        else:
            scale_log(None)

    def cmap_nipyspectral(_):
        nonlocal cmap
        cmap = 'nipy_spectral'
        if scale == 'linear':
            scale_linear(None)
        else:
            scale_log(None)

    ax_button_cmap_hot = plt.axes([0.05, 0.55, 0.1, 0.05])
    button_cmap_hot = Button(ax_button_cmap_hot, 'hot')
    button_cmap_hot.on_clicked(cmap_hot)
    ax_button_nipyspectral = plt.axes([0.05, 0.6, 0.1, 0.05])
    button_nipyspectral = Button(
        ax_button_nipyspectral, 'nipy_spectral')
    button_nipyspectral.on_clicked(cmap_nipyspectral)

    # Figure update
    def update(i):
        nonlocal current_idx
        current_idx = int(i)
        img.set_array(xyz_p[current_idx] + (1 if scale == 'log' else 0))
        text.set_text(f'{focus_slider[current_idx]:.2f} {slider_unit}')

    def update_prev(_):
        slider.set_val(0 if current_idx == 0 else current_idx - 1)

    def update_next(_):
        slider.set_val(max_idx if current_idx == max_idx else current_idx + 1)

    ax_prev = plt.axes([0.15, 0.1, 0.03, 0.03])
    button_prev = Button(ax_prev, '<')
    button_prev.on_clicked(update_prev)
    ax_next = plt.axes([0.8, 0.1, 0.03, 0.03])
    button_next = Button(ax_next, '>')
    button_next.on_clicked(update_next)

    slider.on_changed(update)
    plt.show()


def plot_txy_interactive(data: Union[NLOSCaptureData, NLOSCaptureData.HType],
                         cmap: str = 'hot', slice_axis: str = 't'):
    if isinstance(data, NLOSCaptureData):
        assert data.H_format == HFormat.T_Sx_Sy, \
            'plot_txy_interactive does not support this data format'
        txy = data.H
        delta_t = data.delta_t
        t_start = data.t_start
    else:
        assert data.ndim == 3, \
            'plot_txy_interactive does not support this data format'
        txy = data
        delta_t = None
        t_start = None

    title = 'H(t, x, y) '
    if slice_axis == 't':
        axis = 0
        n_it = txy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += '(T axis slices)'
        slider_title = 'Bins'
        slider_unit = 'index'
        if delta_t is not None and t_start is not None:
            it_v = (t_start + it_v * delta_t) * 1e12 / SPEED_OF_LIGHT
            slider_unit = 'ps'
        xlabel = 'x'
        ylabel = 'y'
    elif slice_axis == 'x':
        axis = 1
        n_it = txy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += '(Y axis slices)'
        slider_title = 'Planes'
        slider_unit = 'index'
        txy = txy.swapaxes(0, 2)
        xlabel = 't'
        ylabel = 'x'
    elif slice_axis == 'y':
        axis = 2
        n_it = txy.shape[axis]
        it_v = np.arange(n_it, dtype=np.float32)
        title += '(X axis slices)'
        slider_title = 'Planes'
        slider_unit = 'index'
        txy = txy.swapaxes(0, 1)
        xlabel = 't'
        ylabel = 'y'
    else:
        raise AssertionError('slice_axis must be one of ("t", "x", "y")')

    return plot_3d_interactive_axis(txy, it_v, axis=axis,
                                    title=title,
                                    slider_title=slider_title,
                                    slider_unit=slider_unit,
                                    cmap=cmap,
                                    xlabel=xlabel, ylabel=ylabel)



"""
tal.reconstruct.fbp
===================

Reconstruction using filtered backprojection.

Shortcut to call reconstruct.filter_H (pf filter) and reconstruct.bp.solve (see respective docstrings for details)
"""

from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
import numpy as np


def solve(data: NLOSCaptureData,
          wl_mean: float,
          wl_sigma: float,
          border: str = 'zero',
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
          camera_system: CameraSystem = CameraSystem.DIRECT_LIGHT,
          projector_focus: NLOSCaptureData.Array3 = None,
          compensate_invsq: bool = False,
          progress: bool = True) -> NLOSCaptureData.SingleReconstructionType:
    """
    See module description of tal.reconstruct.fbp

    data
        See tal.io.read_capture

    wl_mean, wl_sigma, border
        Filter parameters. See tal.reconstruct.filter_H

    volume_xyz
        Multi-dimensional array with XYZ coordinates of the volume voxels.
        See tal.enums.VolumeFormat for possible input formats.
        See e.g. tal.reconstruct.get_volume_min_max_resolution for utilities to generate this array.

    volume_format
        See tal.enums.VolumeFormat

    camera_system
        See tal.enums.CameraSystem

    projector_focus
        Setting that changes how the virtual light is focused onto the hidden volume.
        'None' focuses the virtual light at the same points as the virtual camera (confocal camera)
            This is the behaviour of most imaging algorithms,
            especially the tal.enums.CameraSystem.CONFOCAL_TIME_GATED camera.
        When projector_focus = [x, y, z] if you have multiple laser points in your data,
            the illumination will be focused towards the point [x, y, z].
            This behaviour is especially useful for tal.enums.CameraSystem.PROJECTOR_CAMERA.
        In the pf_dev module you can set projector_focus = volume_xyz and that will yield
            a NLOSCaptureData.ExhaustiveReconstructionType with all possible projector_focus points.

    compensate_invsq
        If True, the inverse square falloff of light is compensated for, i.e., objects further away
        from the relay wall will appear brighter in the reconstruction.

    progress
        If True, shows a progress bar with estimated time remaining.
    """
    from tal.reconstruct import filter_H
    old_H = data.H
    data.H = filter_H(data, filter_name='pf', border=border,
                      wl_mean=wl_mean, wl_sigma=wl_sigma)

    from tal.reconstruct.bp import solve as bp_solve
    H_1 = bp_solve(data,
                   volume_xyz=volume_xyz,
                   volume_format=volume_format,
                   camera_system=camera_system,
                   projector_focus=projector_focus,
                   compensate_invsq=compensate_invsq,
                   progress=progress)
    data.H = old_H
    return H_1



"""
tal.reconstruct.pf_dev
======================

Reconstruction using the phasor fields framework.
See "Non-Line-of-Sight Imaging using Phasor Field Virtual Wave Optics."

This implementation is particularly well suited when:

- The 3D volume that you want to reconstruct is coplanar to the relay wall
and/or
- You want a time-resolved reconstruction, not a time-gated reconstruction
and/or
- You have a projector camera system and want to focus light to multiple points

If that is not your case, you might want to check the tal.reconstruct.pf_dev module.

Does _NOT_ attempt to compensate effects caused by attenuation:
    - cos decay i.e. {sensor|laser}_grid_normals are ignored
    - 1/d^2 decay
"""

from tal.io.capture_data import NLOSCaptureData
from tal.enums import VolumeFormat, CameraSystem
from typing import Union
import numpy as np


def solve(data: NLOSCaptureData,
          wl_mean: float,
          wl_sigma: float,
          border: str = 'zero',
          volume_xyz: NLOSCaptureData.VolumeXYZType = None,
          volume_format: VolumeFormat = VolumeFormat.UNKNOWN,
          camera_system: CameraSystem = CameraSystem.DIRECT_LIGHT,
          projector_focus: Union[NLOSCaptureData.Array3,
                                 NLOSCaptureData.VolumeXYZType] = None,
          progress: bool = True,
          compensate_invsq: bool = False,
          try_optimize_convolutions: bool = True) -> Union[NLOSCaptureData.SingleReconstructionType,
                                                           NLOSCaptureData.ExhaustiveReconstructionType]:
    """
    See module description of tal.reconstruct.pf_dev

    data
        See tal.io.read_capture

    wl_mean, wl_sigma, border
        Filter parameters. See tal.reconstruct.filter_H

    volume_xyz
        Multi-dimensional array with XYZ coordinates of the volume voxels.
        See tal.enums.VolumeFormat for possible input formats.
        See e.g. tal.reconstruct.get_volume_min_max_resolution for utilities to generate this array.

    volume_format
        See tal.enums.VolumeFormat

    camera_system
        See tal.enums.CameraSystem

    projector_focus
        Setting that changes how the virtual light is focused onto the hidden volume.
        'None' focuses the virtual light at the same points as the virtual camera (confocal camera)
            This is the behaviour of most imaging algorithms,
            especially the tal.enums.CameraSystem.CONFOCAL_TIME_GATED camera.
        When projector_focus = [x, y, z] if you have multiple laser points in your data,
            the illumination will be focused towards the point [x, y, z].
            This behaviour is especially useful for tal.enums.CameraSystem.PROJECTOR_CAMERA.
        When projector_focus = volume_xyz, this will yield
            a NLOSCaptureData.ExhaustiveReconstructionType with all possible projector_focus points.

    compensate_invsq
        If True, the inverse square falloff of light is compensated for, i.e., objects further away
        from the relay wall will appear brighter in the reconstruction.

    progress
        If True, shows a progress bar with estimated time remaining.

    try_optimize_convolutions
        When volume_xyz consists of depth-slices (Z-slices) that are 
            1) coplanar to the XY relay wall, and
            2) sampled at the same rate,
        the computation can be optimized to use less memory and be much faster.
        It is recommended to leave this set this to True. If it is not possible to apply
        this optimization, tal.reconstruct.pf_dev will fall back to the default implementation.
        You can generate these depth-slices with tal.reconstruct.get_volumeXXX functions.
    """
    from tal.reconstruct.util import convert_to_N_3, convert_reconstruction_from_N_3
    H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3, \
        optimize_projector_convolutions, optimize_camera_convolutions = \
        convert_to_N_3(data, volume_xyz, volume_format,
                       try_optimize_convolutions=try_optimize_convolutions)

    from tal.reconstruct.pf_dev.phasor_fields import backproject_pf_multi_frequency
    reconstructed_volume_n3 = backproject_pf_multi_frequency(
        H, laser_grid_xyz, sensor_grid_xyz, volume_xyz_n3, volume_xyz.shape[:-1],
        camera_system, data.t_accounts_first_and_last_bounces,
        data.t_start, data.delta_t, data.is_confocal(),
        projector_focus,
        wl_mean, wl_sigma, border,
        optimize_projector_convolutions, optimize_camera_convolutions,
        data.laser_xyz, data.sensor_xyz,
        compensate_invsq=compensate_invsq,
        progress=progress)

    mutliple_projector_points = \
        camera_system.implements_projector() \
        and projector_focus is not None and len(projector_focus) > 3

    return convert_reconstruction_from_N_3(data, reconstructed_volume_n3, volume_xyz, volume_format, camera_system,
                                           is_exhaustive_reconstruction=mutliple_projector_points)



import os
import sys
import shutil
import yaml
import tal
from tal.io.capture_data import NLOSCaptureData
from tal.enums import FileFormat, GridFormat, HFormat
from tal.config import local_file_path
from tal.log import log, LogLevel, TQDMLogRedirect
import datetime
import numpy as np
from tqdm import tqdm
import multiprocessing
from multiprocessing.queues import Queue
from functools import partial

from tal.render.util import import_mitsuba_backend


def render_nlos_scene(config_path, args, num_retries=0):
    config_path = os.path.abspath(config_path)

    assert os.path.exists(config_path), \
        f'{config_path} does not exist'

    path, name = os.path.split(config_path)
    candidate_filename = os.path.join(path, name, f'{name}.yaml')
    if os.path.isdir(config_path) and os.path.isfile(candidate_filename):
        config_path = candidate_filename

    assert os.path.isfile(config_path), \
        f'{config_path} is not a TAL config file'

    config_dir, config_filename = os.path.split(config_path)

    try:
        scene_config = yaml.safe_load(
            open(config_path, 'r')) or dict()

        if scene_config.get('mitsuba_variant', '').startswith('cuda'):
            assert len(args.gpus) > 0, \
                'You must specify at least one GPU to use CUDA. Use tal --gpu <id1> <id2> ...'
            gpu_ids = ','.join(map(str, args.gpus))
            os.environ['CUDA_VISIBLE_DEVICES'] = gpu_ids
        else:
            os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

        mitsuba_backend = import_mitsuba_backend()

        scene_defaults = yaml.safe_load(
            open(local_file_path('render/scene_defaults.yaml'), 'r'))
        scene_defaults['mitsuba_variant'] = mitsuba_backend.get_default_variant()

    except yaml.YAMLError as exc:
        raise AssertionError(
            f'Invalid YAML format in TAL config file: {exc}') from exc
    scene_config = {**scene_defaults, **scene_config}

    if not args.dry_run:
        mitsuba_backend.set_variant(scene_config['mitsuba_variant'])

    steady_xml, nlos_xml = mitsuba_backend.get_scene_xml(
        scene_config, random_seed=args.seed, quiet=args.quiet)

    try:
        in_progress = False
        progress_file = os.path.join(config_dir, 'IN_PROGRESS')
        if os.path.exists(progress_file):
            with open(progress_file, 'r') as f:
                progress_folder = f.read()
            if os.path.exists(os.path.join(config_dir, progress_folder)) and not args.quiet:
                in_progress = True
            else:
                log(LogLevel.INFO, 'The IN_PROGRESS file is stale, removing it...')
                os.remove(progress_file)
        if in_progress and not args.quiet:
            log(LogLevel.INFO,
                f'Found a render in progress ({progress_folder}), continuing...')
        if not in_progress:
            progress_folder = datetime.datetime.now().strftime(r'%Y%m%d-%H%M%S')

        root_dir = os.path.join(config_dir, progress_folder)
        partial_results_dir = os.path.join(root_dir, 'partial')
        steady_dir = os.path.join(root_dir, 'steady')
        log_dir = os.path.join(root_dir, 'logs')

        if not in_progress:
            os.mkdir(root_dir)
            os.mkdir(partial_results_dir)
            os.mkdir(steady_dir)
            os.mkdir(log_dir)
            shutil.copy(
                config_path,
                os.path.join(root_dir, f'{config_filename}.old'))
        with open(progress_file, 'w') as f:
            f.write(progress_folder)
    except OSError as exc:
        raise AssertionError(f'Invalid permissions: {exc}') from exc

    try:
        steady_scene_xml = os.path.join(root_dir, 'steady_scene.xml')
        with open(steady_scene_xml, 'w') as f:
            f.write(steady_xml)

        nlos_scene_xml = os.path.join(root_dir, 'nlos_scene.xml')
        with open(nlos_scene_xml, 'w') as f:
            f.write(nlos_xml)

        laser_lookats = []
        name = scene_config['name']
        scan_type = scene_config['scan_type']
        num_bins = scene_config['num_bins']
        sensor_width = scene_config['sensor_width']
        sensor_height = scene_config['sensor_height']
        laser_width = scene_config['laser_width']
        laser_height = scene_config['laser_height']
        if scan_type == 'single':
            laser_width = 1
            laser_height = 1

        relay_wall = next(filter(
            lambda g: g['name'] == scene_config['relay_wall'],
            scene_config['geometry']))
        assert 'rot_degrees_x' not in relay_wall and \
            'rot_degrees_y' not in relay_wall and \
            'rot_degrees_z' not in relay_wall, \
            'Relay wall displacement/rotation is NYI'

        def get_grid_xyz(nx, ny, rw_scale_x, rw_scale_y):
            px = rw_scale_x
            py = rw_scale_y
            xg = np.stack(
                (np.linspace(-px, px, num=2*nx + 1)[1::2],)*ny, axis=1)
            yg = np.stack(
                (np.linspace(-py, py, num=2*ny + 1)[1::2],)*nx, axis=0)
            assert xg.shape[0] == yg.shape[0] == nx and xg.shape[1] == yg.shape[1] == ny, \
                'Incorrect shapes'
            return np.stack([xg, yg, np.zeros((nx, ny))], axis=-1).astype(np.float32)

        def expand(vec, x, y):
            assert len(vec) == 3
            return vec.reshape(1, 1, 3).repeat(x, axis=0).repeat(y, axis=1)

        if scan_type == 'single':
            laser_lookat_x = \
                scene_config['laser_lookat_x'] or sensor_width / 2
            laser_lookat_y = \
                scene_config['laser_lookat_y'] or sensor_height / 2
            laser_lookats.append((laser_lookat_x, laser_lookat_y))
        elif scan_type == 'exhaustive' or scan_type == 'confocal':
            assert not (scan_type == 'confocal' and
                        (laser_width != sensor_width or
                            laser_height != sensor_height)), \
                'If using scan_type=confocal, sensor_{width|height} must match laser_{width|height}'
            for y in range(laser_height):
                for x in range(laser_width):
                    laser_lookat_x = (x + 0.5) * sensor_width / laser_width
                    laser_lookat_y = (y + 0.5) * sensor_height / laser_height
                    laser_lookats.append((laser_lookat_x, laser_lookat_y))
        else:
            raise AssertionError(
                'Invalid scan_type, must be one of {single|exhaustive|confocal}')

        # TODO(diego): rotate
        displacement = np.array([
            relay_wall['displacement_x'],
            relay_wall['displacement_y'],
            relay_wall['displacement_z']])
        sensor_grid_xyz = get_grid_xyz(
            sensor_width, sensor_height, relay_wall['scale_x'], relay_wall['scale_y'])
        sensor_grid_xyz += displacement
        if scan_type == 'single':
            px = relay_wall['scale_x'] * \
                ((laser_lookat_x / sensor_width) * 2 - 1)
            py = relay_wall['scale_y'] * \
                ((laser_lookat_y / sensor_height) * 2 - 1)
            laser_grid_xyz = np.array([[
                [px, py, 0],
            ]], dtype=np.float32)
        else:
            laser_grid_xyz = get_grid_xyz(
                laser_width, laser_height, relay_wall['scale_x'], relay_wall['scale_y'])
        laser_grid_xyz += displacement
        # TODO(diego): rotate [0, 0, 1] by rot_degrees_x (assmes RW is a plane)
        # or use a more generalist approach
        sensor_grid_normals = expand(
            np.array([0, 0, 1]), sensor_width, sensor_height)
        laser_grid_normals = expand(
            np.array([0, 0, 1]), laser_width, laser_height)

        experiment_name = scene_config['name']

        class StdoutQueue(Queue):
            def __init__(self, *args, **kwargs):
                ctx = multiprocessing.get_context()
                super(StdoutQueue, self).__init__(*args, **kwargs, ctx=ctx)

            def write(self, msg):
                self.put(msg)

        if args.do_steady_renders:
            def render_steady(render_name, sensor_index):
                if not args.quiet:
                    log(LogLevel.INFO,
                        f'{render_name} for {experiment_name} steady render...')
                hdr_ext = mitsuba_backend.get_hdr_extension()
                hdr_path = os.path.join(partial_results_dir,
                                        f'{experiment_name}_{render_name}.{hdr_ext}')
                ldr_path = os.path.join(steady_dir,
                                        f'{experiment_name}_{render_name}.png')
                if os.path.exists(ldr_path) and not args.quiet:
                    pass  # skip
                else:
                    logfile = None
                    if args.do_logging and not args.dry_run:
                        logfile = open(os.path.join(
                            log_dir, f'{experiment_name}_{render_name}.log'), 'w')
                    # NOTE: something here has a memory leak (probably Mitsuba-related)
                    # We run Mitsuba in a separate process to ensure that the leaks do not add up
                    # as they can fill your RAM in exhaustive scans
                    queue = StdoutQueue()
                    run_mitsuba_f = partial(mitsuba_backend.run_mitsuba, steady_scene_xml, hdr_path, dict(),
                                            render_name, logfile, args, sensor_index, queue)
                    if os.name == 'nt':
                        # NOTE: Windows does not support multiprocessing
                        run_mitsuba_f()
                    else:
                        process = multiprocessing.Process(target=run_mitsuba_f)
                        try:
                            process.start()
                            process.join()
                        except KeyboardInterrupt:
                            process.terminate()
                            raise KeyboardInterrupt
                    if args.do_logging and not args.dry_run:
                        while not queue.empty():
                            e = queue.get()
                            if isinstance(e, Exception):
                                raise e
                            else:
                                logfile.write(e)
                        queue.close()
                        logfile.close()
                    if not args.dry_run:
                        mitsuba_backend.convert_hdr_to_ldr(hdr_path, ldr_path)

            render_steady('back_view', 0)
            render_steady('side_view', 1)

        pbar = tqdm(
            enumerate(laser_lookats), desc=f'Rendering {experiment_name} ({scan_type})...',
            file=TQDMLogRedirect(), ascii=True, total=len(laser_lookats))
        for i, (laser_lookat_x, laser_lookat_y) in pbar:
            try:
                hdr_path, is_dir = mitsuba_backend.partial_laser_path(
                    partial_results_dir, experiment_name, laser_lookat_x, laser_lookat_y)
                if os.path.exists(hdr_path) and not args.quiet:
                    continue  # skip
                if is_dir:
                    os.mkdir(hdr_path)
            except OSError as exc:
                raise AssertionError(f'Invalid permissions: {exc}') from exc
            defines = {
                'laser_lookat_x': laser_lookat_x,
                'laser_lookat_y': laser_lookat_y,
            }
            logfile = None
            if args.do_logging and not args.dry_run:
                logfile = open(os.path.join(
                    log_dir,
                    f'{experiment_name}_L[{laser_lookat_x}][{laser_lookat_y}].log'), 'w')
            queue = StdoutQueue()
            render_name = f'Laser {i + 1} of {len(laser_lookats)}'
            # NOTE: something here has a memory leak (probably Mitsuba-related)
            # We run Mitsuba in a separate process to ensure that the leaks do not add up
            # as they can fill your RAM in exhaustive scans
            run_mitsuba_f = partial(mitsuba_backend.run_mitsuba, nlos_scene_xml, hdr_path, defines,
                                    render_name, sys.stdout, args, queue=queue)
            if os.name == 'nt':
                # NOTE: Windows does not support multiprocessing
                run_mitsuba_f()
            else:
                process = multiprocessing.Process(target=run_mitsuba_f)
                try:
                    process.start()
                    process.join()
                except KeyboardInterrupt:
                    process.terminate()
                    raise KeyboardInterrupt
            if scan_type == 'exhaustive' and i == 0:
                size_bytes = os.path.getsize(hdr_path)
                final_size_gb = size_bytes * len(laser_lookats) / 2**30
                pbar.set_description(
                    f'Rendering {experiment_name} ({scan_type}, estimated size: {final_size_gb:.2f} GB)...')
            if args.do_logging and not args.dry_run:
                past_elems = []
                while not queue.empty():
                    e = queue.get()
                    past_elems.append(e)
                    if isinstance(e, Exception):
                        log(LogLevel.ERROR, '')
                        log(LogLevel.ERROR, '/!\ Mitsuba thread got an exception!')
                        log(LogLevel.ERROR, '')
                        for pe in past_elems[:-1]:
                            log(LogLevel.ERROR, pe)
                        log(LogLevel.ERROR, '')
                        raise e
                    else:
                        logfile.write(e)
                    logfile.flush()
                queue.close()
                logfile.close()

        if args.dry_run:
            return

        if not args.quiet:
            log(LogLevel.INFO, 'Merging partial results...')

        capture_data = NLOSCaptureData()
        capture_data.sensor_xyz = np.array([
            scene_config['sensor_x'],
            scene_config['sensor_y'],
            scene_config['sensor_z'],
        ], dtype=np.float32)
        capture_data.sensor_grid_xyz = sensor_grid_xyz
        capture_data.sensor_grid_normals = sensor_grid_normals
        capture_data.sensor_grid_format = GridFormat.X_Y_3
        capture_data.laser_xyz = np.array([
            scene_config['laser_x'],
            scene_config['laser_y'],
            scene_config['laser_z'],
        ], dtype=np.float32)
        capture_data.laser_grid_xyz = laser_grid_xyz
        capture_data.laser_grid_normals = laser_grid_normals
        capture_data.laser_grid_format = GridFormat.X_Y_3
        # NOTE(diego): we do not store volume information for now
        # capture_data.volume_format = VolumeFormat.X_Y_Z_3
        capture_data.delta_t = scene_config['bin_width_opl']
        capture_data.t_start = scene_config['start_opl']
        capture_data.t_accounts_first_and_last_bounces = \
            scene_config['account_first_and_last_bounces']
        capture_data.scene_info = {
            'tal_version': tal.__version__,
            'config': scene_config,
            'args': vars(args),
        }
        if scan_type == 'single':
            hdr_path, _ = mitsuba_backend.partial_laser_path(
                partial_results_dir,
                experiment_name,
                *laser_lookats[0])
            capture_data.H = mitsuba_backend.read_transient_image(hdr_path)
            capture_data.H_format = HFormat.T_Sx_Sy
        elif scan_type == 'exhaustive' or scan_type == 'confocal':
            if scan_type == 'exhaustive':
                capture_data.H = np.empty(
                    (num_bins, laser_width, laser_height,
                     sensor_width, sensor_height),
                    dtype=np.float32)
                capture_data.H_format = HFormat.T_Lx_Ly_Sx_Sy
            elif scan_type == 'confocal':
                capture_data.H = np.empty(
                    (num_bins, laser_width, laser_height),
                    dtype=np.float32)
                capture_data.H_format = HFormat.T_Sx_Sy
            else:
                raise AssertionError

            e_laser_lookats = enumerate(laser_lookats)
            if not args.quiet and len(laser_lookats) > 1:
                e_laser_lookats = tqdm(
                    e_laser_lookats, desc='Merging partial results...',
                    file=TQDMLogRedirect(), ascii=True, total=len(laser_lookats))
            try:
                for i, (laser_lookat_x, laser_lookat_y) in e_laser_lookats:
                    x = i % laser_width
                    y = i // laser_width
                    hdr_path, _ = mitsuba_backend.partial_laser_path(
                        partial_results_dir,
                        experiment_name,
                        laser_lookat_x, laser_lookat_y)
                    if scan_type == 'confocal':
                        capture_data.H[:, x:x+1, y:y+1, ...] = \
                            mitsuba_backend.read_transient_image(hdr_path)
                    elif scan_type == 'exhaustive':
                        capture_data.H[:, x, y, ...] = \
                            mitsuba_backend.read_transient_image(hdr_path)
                    else:
                        raise AssertionError
            except Exception as exc:
                if num_retries >= 10:
                    raise AssertionError(
                        f'Failed to read partial results after {num_retries} retries')
                mitsuba_backend.remove_transient_image(hdr_path)
                # TODO Mitsuba sometimes fails to write some images,
                # it seems like some sort of race condition
                # If there is a partial result missing, just re-launch for now
                log(LogLevel.INFO,
                    f'We missed some partial results (iteration {i} failed because: {exc}), re-launching...')
                return render_nlos_scene(config_path, args, num_retries=num_retries + 1)
        else:
            raise AssertionError(
                'Invalid scan_type, must be one of {single|exhaustive|confocal}')

        hdf5_path = os.path.join(root_dir, f'{experiment_name}.hdf5')
        tal.io.write_capture(capture_data, hdf5_path,
                             file_format=FileFormat.HDF5_TAL)

        if not args.quiet:
            log(LogLevel.INFO, f'Stored result in {hdf5_path}')

        # remove IN_PROGRESS file
        os.remove(progress_file)

        if args.keep_partial_results:
            return

        if not args.quiet:
            log(LogLevel.INFO,
                f'Cleaning partial results in {partial_results_dir}...')

        shutil.rmtree(partial_results_dir)

        if not args.quiet:
            log(LogLevel.INFO, f'All clean.')
    except KeyboardInterrupt:
        delete = None
        while delete is None:
            try:
                log(LogLevel.PROMPT,
                    f'Render cancelled. '
                    f'Delete the directory {root_dir}? (y/n): ', end='')
                answer = input()
                if answer.lower() == 'y':
                    delete = True
                elif answer.lower() == 'n':
                    delete = False
            except KeyboardInterrupt:
                pass
        if delete:
            shutil.rmtree(root_dir)
            os.remove(progress_file)



import numpy as np
import tal
from tal.config import get_resources
from tal.log import log, LogLevel, TQDMLogRedirect
from tqdm import tqdm


def backproject_pf_multi_frequency(
        H_0, laser_grid_xyz, sensor_grid_xyz, volume_xyz, volume_xyz_shape,
        camera_system, t_accounts_first_and_last_bounces,
        t_start, delta_t, is_confocal,
        projector_focus,
        frequencies, weights, nt,
        optimize_projector_convolutions, optimize_camera_convolutions,
        laser_xyz=None, sensor_xyz=None, progress=False):

    assert not is_confocal, 'tal.reconstruct.upf_dev does not support confocal captures. ' \
        'Please use tal.reconstruct.bp or tal.reconstruct.fbp instead.'

    if not optimize_projector_convolutions and not optimize_camera_convolutions and not camera_system.is_transient():
        log(LogLevel.WARNING, 'tal.reconstruct.upf_dev: You have specified a time-gated camera system '
            'with an arbitrary reconstruction volume (that is not parallel to the relay wall). '
            'This will work, but the tal.reconstruct.bp or tal.reconstruct.fbp implementations '
            'are better suited for these cases.')

    nw, nl, ns = H_0.shape
    nv = np.prod(volume_xyz.shape[:-1])  # N or X * Y or X * Y * Z
    if optimize_projector_convolutions or optimize_camera_convolutions:
        if volume_xyz.ndim == 3:
            nvx, nvy, _ = volume_xyz.shape
            volume_xyz = volume_xyz.reshape((nvx, nvy, 1, 3))
        assert volume_xyz.ndim == 4, 'Expecting X_Y_Z_3 format'
        nvx, nvy, nvz, _ = volume_xyz.shape
    if optimize_projector_convolutions:
        assert laser_grid_xyz.ndim == 3, 'Expecting X_Y_3 format'
        nlx, nly, _ = laser_grid_xyz.shape
    else:
        nlx, nly = 1, 1
    if optimize_camera_convolutions:
        assert sensor_grid_xyz.ndim == 3, 'Expecting X_Y_3 format'
        nsx, nsy, _ = sensor_grid_xyz.shape
    else:
        nsx, nsy = 1, 1

    """ Phasor fields filter """

    if weights is None:
        weights = np.ones(nw, dtype=np.float32, dtype=np.float32)
    else:
        weights = weights.astype(np.complex64)
    freqs = frequencies.astype(np.float32)
    assert len(freqs) == len(weights) == nw, \
        'len(freqs), len(weights), and H_0.shape[0] are not the same.'

    t = np.arange(nt, dtype=np.float32) * delta_t
    padding = 0
    # min_freq = freqs[~np.isclose(freqs, 0)].min()
    # nt = int(np.ceil((1 / min_freq) / delta_t))
    # log(LogLevel.INFO, f'tal.reconstruct.upf_dev: computed nt is {nt}')

    """ Distance calculation and propagation """

    if camera_system.implements_projector():
        assert projector_focus is not None, \
            'projector_focus is required for this camera system'
        if len(projector_focus) == 3:
            projector_focus = np.array(projector_focus).reshape(
                (1, 1, 1, 3))
            projector_focus_mode = 'single'
            if optimize_projector_convolutions:
                log(LogLevel.INFO, 'tal.reconstruct.upf_dev: When projector_focus is a 3D point, the projector convolution optimization is not implemented. '
                    'Falling back to default method.')
            optimize_projector_convolutions = False
        else:
            assert np.allclose(projector_focus.flatten(), volume_xyz.flatten()), \
                'projector_focus must be a single 3D point, ' \
                'or you should pass the same value as volume_xyz.'
            assert nvz == 1, \
                'When projector_focus=volume_xyz, the volume must be a single Z slice. ' \
                f'In your case, your volume has {nvz} Z slices. ' \
                'You can call this same function for each individual point in the volume.'
            projector_focus = projector_focus.reshape((1, nv, 1, 3))
            projector_focus_mode = 'exhaustive'
    else:
        assert projector_focus is None, \
            'projector_focus must not be set for this camera system'
        projector_focus = volume_xyz.reshape((1, nv, 1, 3))
        projector_focus_mode = 'confocal'
        if optimize_projector_convolutions:
            log(LogLevel.INFO, 'tal.reconstruct.upf_dev: When projector_focus is not set, the projector convolution optimization is not implemented. '
                'Falling back to default method.')
        optimize_projector_convolutions = False

    log(LogLevel.INFO, 'tal.reconstruct.upf_dev: '
        f'projector_focus_mode={projector_focus_mode}')
    projector_focus = projector_focus.astype(np.float32)

    # reshape everything into (nl, nv, ns, 3)
    if laser_xyz is not None:
        laser_xyz = laser_xyz.reshape((1, 1, 1, 3)).astype(np.float32)
    if sensor_xyz is not None:
        sensor_xyz = sensor_xyz.reshape((1, 1, 1, 3)).astype(np.float32)
    laser_grid_xyz = laser_grid_xyz.reshape((nl, 1, 1, 3)).astype(np.float32)
    sensor_grid_xyz = sensor_grid_xyz.reshape((1, 1, ns, 3)).astype(np.float32)
    volume_xyz = volume_xyz.reshape((1, nv, 1, 3)).astype(np.float32)

    def distance(a, b):
        return np.linalg.norm(b - a, axis=-1)

    # d_0: t_start (moment the sensor starts capturing w.r.t. pulse emission)
    # d_1: laser origin to laser illuminated point
    # d_2: laser illuminated point to projector_focus
    # d_3: x_v (camera_focus) to sensor imaged point
    # d_4: sensor imaged point to sensor origin
    d_0 = np.float32(t_start * -1)

    if t_accounts_first_and_last_bounces:
        d_1 = distance(laser_xyz, laser_grid_xyz).reshape((nl, 1))
        d_4 = distance(sensor_grid_xyz, sensor_xyz).reshape((1, ns))
    else:
        d_1 = np.float32(0.0)
        d_4 = np.float32(0.0)

    if projector_focus_mode == 'exhaustive':
        n_projector_points = nv
    else:
        n_projector_points = 1

    d_014 = d_0 + d_1 + d_4

    if optimize_projector_convolutions or optimize_camera_convolutions:
        nvi = nvx * nvy
        range_z = range(nvz)
    else:
        nvi = nv
        range_z = [0]
        nvz = 1
    range_z = enumerate(range_z)
    if nvz > 1 and progress:
        range_z = tqdm(
            range_z,
            desc='tal.reconstruct.pf_dev Z slices',
            total=nvz,
            file=TQDMLogRedirect(),
            position=0,
            leave=True)

    if camera_system.is_transient():
        H_1 = np.zeros(
            (nt, n_projector_points, nvi, nvz),
            dtype=np.complex64)
    else:
        H_1 = np.zeros(
            (n_projector_points, nvi, nvz),
            dtype=np.complex64)

    for i_z, nvzi in range_z:
        if optimize_projector_convolutions:
            projector_focus_i = projector_focus.reshape(
                (nvx, nvy, nvz, 3))[..., nvzi, :]
        else:
            projector_focus_i = projector_focus
        if optimize_camera_convolutions:
            volume_xyz_i = volume_xyz.reshape(
                (nvx, nvy, nvz, 3))[..., nvzi, :]
        else:
            volume_xyz_i = volume_xyz

        if camera_system.bp_accounts_for_d_2():
            if optimize_projector_convolutions:
                assert projector_focus_mode in ['exhaustive', 'confocal']
                rlx = nvx + nlx - 1
                rly = nvy + nly - 1
                laser_grid_xyz = laser_grid_xyz.reshape((nlx, nly, 3))
                l_dx = laser_grid_xyz[1, 0, ...] - \
                    laser_grid_xyz[0, 0, ...]
                l_dy = laser_grid_xyz[0, 1, ...] - \
                    laser_grid_xyz[0, 0, ...]

                p0 = projector_focus_i[0, 0, ...] - \
                    laser_grid_xyz[-1, -1, ...]

                i_vals, j_vals = np.meshgrid(
                    np.arange(rlx), np.arange(rly), indexing='ij')
                i_vals = i_vals.reshape((rlx, rly, 1))
                j_vals = j_vals.reshape((rlx, rly, 1))
                p0 = p0.reshape((1, 1, 3))
                l_dx = l_dx.reshape((1, 1, 3))
                l_dy = l_dy.reshape((1, 1, 3))
                d_2 = np.linalg.norm(
                    p0 + l_dx * i_vals + l_dy * j_vals, axis=-1).astype(np.float32)
                d_2 = np.fft.ifftshift(d_2)
            elif projector_focus_mode == 'confocal':
                d_2 = distance(
                    laser_grid_xyz, projector_focus_i.reshape((1, nvi, 1, 3)))
            else:
                assert projector_focus_mode == 'single'
                d_2 = distance(
                    laser_grid_xyz, projector_focus_i.reshape((1, 1, 1, 3)).repeat(nvi, axis=1))
        else:
            d_2 = np.float32(0.0)

        if optimize_camera_convolutions:
            rsx = nvx + nsx - 1
            rsy = nvy + nsy - 1
            sensor_grid_xyz = sensor_grid_xyz.reshape((nsx, nsy, 3))
            s_dx = sensor_grid_xyz[1, 0, ...] - \
                sensor_grid_xyz[0, 0, ...]
            s_dy = sensor_grid_xyz[0, 1, ...] - \
                sensor_grid_xyz[0, 0, ...]

            p0 = volume_xyz_i[0, 0, ...] - \
                sensor_grid_xyz[-1, -1, ...]

            i_vals, j_vals = np.meshgrid(
                np.arange(rsx), np.arange(rsy), indexing='ij')
            i_vals = i_vals.reshape((rsx, rsy, 1))
            j_vals = j_vals.reshape((rsx, rsy, 1))
            p0 = p0.reshape((1, 1, 3))
            s_dx = s_dx.reshape((1, 1, 3))
            s_dy = s_dy.reshape((1, 1, 3))
            d_3 = np.linalg.norm(
                p0 + s_dx * i_vals + s_dy * j_vals, axis=-1).astype(np.float32)
            d_3 = np.fft.ifftshift(d_3)
        else:
            d_3 = distance(volume_xyz_i, sensor_grid_xyz)

        nw_pow2 = 2 ** np.ceil(np.log2(nw)).astype(np.int32)
        freqs_pad = np.pad(freqs, (0, nw_pow2 - nw), 'constant')
        weights_pad = np.pad(weights, (0, nw_pow2 - nw), 'constant')

        def work_zslice_freq(range_w):
            nwi = len(range_w)
            freqs_i = freqs_pad[range_w]
            weights_i = weights_pad[range_w]
            H_1_w = np.zeros((nwi, n_projector_points, nvi),
                             dtype=np.complex64)

            fw_iterator = enumerate(zip(freqs_i, weights_i))
            if progress:
                fw_iterator = tqdm(fw_iterator,
                                   desc='tal.reconstruct.upf_dev propagation (1/2)',
                                   file=TQDMLogRedirect(),
                                   total=min(len(freqs_i), nw),
                                   position=0,
                                   leave=True)

            for i_w, (frequency, weight) in fw_iterator:
                if weight == 0.0:
                    # H_1_w[i_w, ...] = 0
                    continue

                H_0_w = H_0[range_w[i_w], ...].reshape((nl, ns))

                rsd_014 = np.exp(np.complex64(2j * np.pi) * d_014 * frequency)
                H_0_w *= rsd_014
                del rsd_014

                rsd_3 = np.exp(np.complex64(2j * np.pi) * d_3 * frequency)
                if optimize_camera_convolutions:
                    H_0_w = H_0_w.reshape((nlx, nly, nsx, nsy))
                    rsd_3 = rsd_3.reshape((1, 1, rsx, rsy))
                    H_0_w_fft = np.fft.fft2(
                        H_0_w, axes=(2, 3), s=(rsx, rsy))
                    rsd_3_fft = np.fft.fft2(
                        rsd_3, axes=(2, 3), s=(rsx, rsy))
                    H_0_w_fft *= rsd_3_fft
                    H_0_w = np.fft.ifft2(H_0_w_fft, axes=(2, 3))
                    H_0_w = H_0_w[:, :, :nvx, :nvy]
                    H_0_w = H_0_w.reshape((nl, nvi))
                    del rsd_3
                else:
                    H_0_w = H_0_w.reshape((nl, 1, ns))
                    H_0_w = H_0_w * rsd_3.reshape((1, nvi, ns))
                    H_0_w = H_0_w.sum(axis=2)

                if camera_system.bp_accounts_for_d_2():
                    rsd_2 = np.exp(np.complex64(2j * np.pi) * d_2 * frequency)
                    if projector_focus_mode == 'exhaustive':
                        assert optimize_projector_convolutions, \
                            'You must use the convolutions optimization when projector_focus=volume_xyz. ' \
                            'Check the documentation for tal.reconstruct.upf_dev for more information.'
                        H_0_w = H_0_w.reshape((nlx, nly, nvx, nvy))
                        rsd_2 = rsd_2.reshape((rlx, rly, 1, 1))
                        H_0_w_fft = np.fft.fft2(
                            H_0_w, axes=(0, 1), s=(rlx, rly))
                        rsd_2_fft = np.fft.fft2(
                            rsd_2, axes=(0, 1), s=(rlx, rly))
                        H_0_w_fft *= rsd_2_fft
                        H_0_w = np.fft.ifft2(H_0_w_fft, axes=(0, 1))
                        H_0_w = H_0_w[:nvx, :nvy, :, :]
                    else:
                        assert projector_focus_mode in ['single', 'confocal']
                        H_0_w = H_0_w.reshape((nl, nvi))
                        H_0_w *= rsd_2.reshape((nl, nvi))
                        H_0_w = H_0_w.sum(axis=0)
                    H_0_w = H_0_w.reshape((n_projector_points, nvi))
                    del rsd_2

                H_1_w[i_w, ...] = weight * H_0_w
                del H_0_w

            return H_1_w

        H_1_w = np.zeros((nw_pow2, n_projector_points, nvi),
                         dtype=np.complex64)
        range_w = np.arange(nw_pow2, dtype=np.int32)

        get_resources().split_work(
            work_zslice_freq,
            data_in=range_w,
            data_out=H_1_w,
            slice_dims=(0, 0),
        )

        # H_1 has shape (n_projector_points, nvi, nvz) or (nt, n_projector_points, nvi, nvz)
        # H_1_w has shape (nw_pow2, n_projector_points, nvi)
        f_iterator = enumerate(freqs)
        if progress:
            f_iterator = tqdm(f_iterator,
                              desc='tal.reconstruct.upf_dev ifft (2/2)',
                              file=TQDMLogRedirect(),
                              total=nw,
                              position=0,
                              leave=True)
        for i_w, frequency in f_iterator:
            e = np.exp(np.complex64(2j * np.pi) * t * frequency) / nt
            H_1_t = (
                e.reshape((nt, 1, 1))
                *
                H_1_w[i_w, ...].reshape((1, n_projector_points, nvi))
            )

            if camera_system.is_transient():
                H_1[..., i_z] += H_1_t[padding:-padding, ...]
            else:
                H_1[..., i_z] += H_1_t[padding, ...]

    if n_projector_points == 1:
        # squeeze n_projector_points axis
        H_1 = H_1.squeeze(axis=-3)

    return H_1



from enum import Enum


class LogLevel(Enum):
    """
    LogLevel, from more verbosity to less verbosity.
    """
    TRACE = 0  # extra debug
    DEBUG = 1
    INFO = 2
    PROGRESS = 3  # progress bars
    WARNING = 4
    ERROR = 5
    PROMPT = 6  # only input() required by the user
    NONE = 7  # no log


_log_level = LogLevel.TRACE


def _set_default_log_level():
    from tal.config import ask_for_config, Config
    global _log_level
    _log_level = LogLevel[ask_for_config(Config.LOG_LEVEL, force_ask=False)]


def set_log_level(level: LogLevel):
    """
    Controls the amount of information that will be logged to the console.

    See tal.LogLevel for the different levels of verbosity.
    """
    global _log_level
    _log_level = level
    from tal.config import Config, read_config, write_config
    config = read_config()
    config[Config.LOG_LEVEL.value[0]] = level.name
    write_config(config)


def log(level: LogLevel, message: str, **kwargs):
    """
    Logs a message to the console.
    """
    if level.value >= _log_level.value:
        # TODO add pretty colors (see libcpp-common) :^)
        print(message, **kwargs)


def TQDMLogRedirect():
    import sys
    if LogLevel.PROGRESS.value >= _log_level.value:
        return sys.stderr
    else:
        return EmptyLogRedirect()


class EmptyLogRedirect:
    def __init__(self):
        pass

    def write(self, text):
        pass

    def flush(self):
        pass

    def getvalue(self):
        return ''
